#include "records.h"

#include <chrono>
#include <print>
#include <unordered_map>
#include <unordered_set>

#include <leveldb/db.h>

#include <nlohmann/json.hpp>

using leveldb_ptr = leveldb::DB*;

auto records_db     = leveldb_ptr{nullptr};
auto records_count_ = 0;

auto records = std::unordered_map<std::string, std::unordered_set<int>>{};

extern "C"
{
    int open_records_db()
    {
        auto opts = leveldb::Options{};

        opts.create_if_missing = true;

        auto status = leveldb::DB::Open(opts, "db/records", &records_db);

        if (!status.ok()) {
            std::println(stderr, "Failed to open records database: {}", status.ToString());
            return 0;
        }

        auto user_id_buf    = std::string{};
        auto problem_id_buf = 0;
        auto status_buf     = std::string{};

        auto it = records_db->NewIterator(leveldb::ReadOptions());
        for (it->SeekToFirst(); it->Valid(); it->Next()) {
            auto json = nlohmann::json::parse(it->value().ToString());
            // std::println(stderr, "{}", json.dump());
            json["data"]["status"].get_to(status_buf);
            if (status_buf == "ac") {
                // std::println(stderr, "{}", json.dump());
                json["user_id"].get_to(user_id_buf);
                json["problem"].get_to(problem_id_buf);
                records[user_id_buf].emplace(problem_id_buf);
            }
            records_count_++;
        }

        if (!it->status().ok()) {
            std::println(stderr, "Failed to count records: {}", it->status().ToString());
            delete it;
            delete records_db;
            return 0;
        }

        delete it;
        return 1;
    }

    void close_records_db()
    {
        delete records_db;
    }

    int add_record(const char* user_id, int problems_id, const char* data)
    {
        auto json = nlohmann::json{
            {"user_id", user_id                    },
            {"problem", problems_id                },
            {"data",    nlohmann::json::parse(data)}
        };
        auto status = records_db->Put(leveldb::WriteOptions{}, std::to_string(records_count_), json.dump());
        if (!status.ok()) {
            std::println(stderr, "Failed to add a record: {}", status.ToString());
            return 0;
        }
        ++records_count_;
        return 1;
    }

    int add_record_impl(const char* user_id, int problems_id, const char* answer, auto status)
    {
        auto now          = std::chrono::utc_clock::now();
        auto timestamp_ms = std::chrono::time_point_cast<std::chrono::milliseconds>(now);
        auto timestamp    = timestamp_ms.time_since_epoch().count();

        auto json = nlohmann::json{
            {"answer",    answer   },
            {"status",    status   },
            {"timestamp", timestamp}
        };
        return add_record(user_id, problems_id, json.dump().c_str());
    }

    int add_record_ac(const char* user_id, int problems_id, const char* answer)
    {
        return add_record_impl(user_id, problems_id, answer, "ac");
    }
    int add_record_wa(const char* user_id, int problems_id, const char* answer)
    {
        return add_record_impl(user_id, problems_id, answer, "wa");
    }
    int add_record_uke(const char* user_id, int problems_id, const char* answer)
    {
        return add_record_impl(user_id, problems_id, answer, "uke");
    }

    int get_record(int id, char** result)
    {
        auto value = std::string{};

        auto status = records_db->Get(leveldb::ReadOptions{}, std::to_string(id), &value);
        if (status.IsNotFound()) {
            *result = NULL;
        } else if (!status.ok()) {
            std::println(stderr, "Failed to get a record: {}", status.ToString());
            return 0;
        } else {
            *result = strdup(value.c_str());
        }
        return 1;
    }

    int all_records_by(const char* user_id, char** result)
    {
        auto problems = std::vector<int>{};
        auto buf      = std::string{};

        auto it = records_db->NewIterator(leveldb::ReadOptions());
        for (it->SeekToFirst(); it->Valid(); it->Next()) {
            auto json = nlohmann::json::parse(it->value().ToString());
            json["user_id"].get_to(buf);
            if (buf != user_id) continue;
            problems.emplace_back(std::stoi(it->key().ToString()));
        }

        if (!it->status().ok()) {
            std::println(stderr, "Failed to get all records by user id: {}", it->status().ToString());
            delete it;
            return 0;
        }
        delete it;
        auto json = nlohmann::json(problems);

        *result = strdup(json.dump().c_str());
        return 1;
    }

    int all_records_of(int problems_id, char** result)
    {
        auto problems = std::vector<int>{};
        auto buf      = 0;

        auto it = records_db->NewIterator(leveldb::ReadOptions());
        for (it->SeekToFirst(); it->Valid(); it->Next()) {
            auto json = nlohmann::json::parse(it->value().ToString());
            json["problem"].get_to(buf);
            if (buf != problems_id) continue;
            problems.emplace_back(std::stoi(it->key().ToString()));
        }

        if (!it->status().ok()) {
            std::println(stderr, "Failed to get all records of the problem: {}", it->status().ToString());
            delete it;
            return 0;
        }
        delete it;
        auto json = nlohmann::json(problems);

        *result = strdup(json.dump().c_str());
        return 1;
    }

    int records_count()
    {
        return records_count_;
    }

    int have_ac(const char* user_id, int problem_id)
    {
        return records[user_id].contains(problem_id);
    }
}